# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

"""
Initializes an empty .rfgb directory for storing models and metadata.
"""

from __future__ import print_function

import os
import sys


def init(quiet=False):

    if not os.path.exists(".rfgb"):
        os.makedirs(".rfgb")
        os.makedirs(".rfgb/models/")

        if not quiet:
            print(
                "Initialized empty rfgb repository at", os.path.abspath(".") + "/.rfgb/"
            )
    else:
        print(".rfgb/ already exists.", file=sys.stderr)
        exit(1)



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

"""
Data structures and methods for learning decision trees.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from .utils import Utils
from .logic import Logic
from .logic import Prover

from copy import deepcopy


class node:
    """
    A node in a tree.

    :param expandQueue: Breadth first search node expansion strategy
    :param depth: initial depth is 0 because no node present
    :param maxDepth: max depth set to 1 because we want to at least learn a tree of depth 1
    :param learnedDecisionTree: this will hold all the clauses learned
    :param data: stores all the facts, positive and negative examples
    """

    expandQueue = []
    depth = 0
    maxDepth = 1
    learnedDecisionTree = []
    data = None

    def __init__(
        self,
        test=None,
        examples=None,
        information=None,
        level=None,
        parent=None,
        pos=None,
    ):
        """
        Constructor for node class.

        :param test: Test condition in the form of a horn clause.
        :param examples: Examples available for testing at this node.
        :param information: Information contained at this node.
        :param level: Level of this node in the tree (0 for root).
        :param parent: "root", or a pointer to the parent.
        :param pos: Position in the tree ('left' or 'right')
        """
        self.test = test
        if level > 0:
            self.parent = parent
        else:
            self.parent = "root"
        self.pos = pos
        self.examples = examples
        self.information = information
        self.level = level
        self.left = None
        self.right = None

        # Add to the queue of nodes to expand.
        node.expandQueue.insert(0, self)

    @staticmethod
    def setMaxDepth(depth):
        """
        Set the maximum depth of the tree.
        """
        node.maxDepth = depth

    @staticmethod
    def initTree(trainingData):
        """
        Create the root node of the tree.
        """

        node.data = trainingData

        # Reset node queue for every tree to be learned.
        node.expandQueue = []
        # Reset clauses for every tree to be learned.
        node.learnedDecisionTree = []

        if trainingData.regression:
            # Regression examples can be collected from trainingData.examples
            # (since there are no pos/neg).
            examples = trainingData.examples.keys()
        else:
            # For all other models, we consider a set of positive and
            # negative examples.
            examples = list(trainingData.pos.keys()) + list(trainingData.neg.keys())

        node(
            test=None,
            examples=examples,
            information=trainingData.variance(examples),
            level=0,
            parent="root",
        )

    @staticmethod
    def learnTree(data):
        """
        Method to create and learn the decision tree.
        """

        # Create the root
        node.initTree(data)

        while len(node.expandQueue) > 0:
            current = node.expandQueue.pop()
            current.expandOnBestTest(data)

        node.learnedDecisionTree.sort(key=lambda x: len(x.split(" ")[0]))
        node.learnedDecisionTree = node.learnedDecisionTree[::-1]

    def getTrueExamples(self, clause, test, data):
        """
        Returns all examples that satisfy the clause
        with conjoined test literal.
        """

        # Initialize a list of true examples.
        trueExamples = []
        clauseCopy = deepcopy(clause)

        # Construct clause for prover
        if clauseCopy[-1] == "-":
            clauseCopy += test
        elif clauseCopy[-1] == ";":
            clauseCopy = clauseCopy.replace(";", ",") + test

        # Prove if example satisfies clause.
        for example in self.examples:
            if Prover.prove(data, example, clauseCopy):
                trueExamples.append(example)
        return trueExamples

    def expandOnBestTest(self, data=None):
        """
        Expand the node based on the best test.
        """

        target = data.getTarget()
        # Initialize clause learned at this node with empty body.
        clause = target + ":-"

        current = self
        ancestorTests = []
        while current.parent != "root":

            if current.pos == "left":
                clause += current.parent.test + ";"
                ancestorTests.append(current.parent.test)
            elif current.pos == "right":
                ancestorTests.append(current.parent.test)

            current = current.parent

        if self.level == node.maxDepth or round(self.information, 3) == 0:

            if clause[-1] != "-":
                node.learnedDecisionTree.append(
                    clause[:-1] + " " + str(Utils.getleafValue(self.examples))
                )
            else:
                node.learnedDecisionTree.append(
                    clause + " " + str(Utils.getleafValue(self.examples))
                )
            return

        if clause[-2] == "-":
            clause = clause[:-1]

        # Initialize minimum weighted variance to a low value.
        minScore = float("inf")
        bestTest = ""

        # List for best test  examples which satisfy  or do not satisfy clause.
        bestTExamples, bestFExamples = [], []
        # Get all the literals contained in the facts.
        literals = data.getLiterals()

        tests = []
        # For every literal generate test conditions.
        for literal in literals:
            literalName = literal[0]
            literalTypeSpecification = literal[1]

            # Generate all possible literal, variable, and constant combinations
            tests += Logic.generateTests(literalName, literalTypeSpecification, clause)

        if self.parent != "root":
            tests = [test for test in tests if not test in ancestorTests]
        tests = set(tests)

        # Check which test scores the best.
        for test in tests:
            # Examples which are satisfied.
            tExamples = self.getTrueExamples(clause, test, data)
            # Examples which are not satisfied (under closed world assumption).
            fExamples = [
                example for example in self.examples if example not in tExamples
            ]
            # Total number of examples.
            example_len = len(self.examples)

            # Calculated the weighted variance:
            score = (len(tExamples) / example_len) * data.variance(tExamples) + (
                len(fExamples) / example_len
            ) * data.variance(fExamples)

            if score < minScore:  # if score lower than current lowest
                minScore = score  # assign new minimum
                bestTest = test  # assign new best test
                bestTExamples = tExamples  # collect satisfied examples
                bestFExamples = fExamples  # collect unsatisfied examples
        Utils.addVariableTypes(bestTest)  # add variable types of new variables
        self.test = bestTest  # assign best test after going through all literal specs

        print("Best test found at the current node: ", self.test)

        # If True examples need further explaining,
        # create left node and add to the queue.
        if len(bestTExamples) > 0:

            self.left = node(
                test=None,
                examples=bestTExamples,
                information=data.variance(bestTExamples),
                level=self.level + 1,
                parent=self,
                pos="left",
            )

            if self.level + 1 > node.depth:
                node.depth = self.level + 1

        # If False examples need further explaining,
        # create right node and add to the queue.
        if len(bestFExamples) > 0:

            self.right = node(
                test=None,
                examples=bestFExamples,
                information=data.variance(bestFExamples),
                level=self.level + 1,
                parent=self,
                pos="right",
            )

            if self.level + 1 > node.depth:
                node.depth = self.level + 1

        # If there are no examples, append clause as it is.
        # if no examples append clause as is
        if self.test == "" or round(self.information, 3) == 0:
            if clause[-1] != "-":
                node.learnedDecisionTree.append(clause[:-1])
            else:
                node.learnedDecisionTree.append(clause)
            return



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

from __future__ import print_function
from __future__ import absolute_import

from ..boosting import updateGradients
from ..tree import node
from ..utils import Utils


def learn(
    targets,
    numTrees=10,
    path="",
    regression=False,
    advice=False,
    softm=False,
    alpha=0.0,
    beta=0.0,
    saveJson=True,
):
    """
    .. versionadded:: 0.3.0

    Learn a relational dependency network from facts and positive/negative
    examples via relational regression trees.

    .. note:: This currently requires that training data is stored as files
              on disk.

    :param targets: List of target predicates to learn models for.
    :type targets: list of str.

    :param numTrees: Number of trees to learn.
    :type numTrees: int.

    :param path: Path to the location training data is stored.
    :type path: str.

    :param regression: Learn a regression model instead of classification.
    :type regression: bool.

    :param advice: Read an advice file from the same directory as trainPath.
    :type advice: bool.

    :default regression: False
    :default advice: False

    :returns: Dictionary where the key is the target and the value is the
              set of trees returned for that target.
    :rtype: dict.
    """

    # Models will be returned as a dictionary, where the name of the predicate
    # will be bound to the set of trees learned for it.
    models = {}

    for target in targets:

        # Read the training data.
        trainData = Utils.readTrainingData(
            target,
            path=path,
            regression=regression,
            advice=advice,
            softm=softm,
            alpha=alpha,
            beta=beta,
        )

        # Initialize an empty list for the trees.
        trees = []

        # Learn each tree and update the gradients.
        for i in range(numTrees):

            node.setMaxDepth(2)
            node.learnTree(trainData)
            trees.append(node.learnedDecisionTree)
            updateGradients(trainData, trees)

            # Save the models learned at this step.
            if saveJson:

                # Collect the parameters used to learn these trees:
                params = {
                    "target": target,
                    "trees": i + 1,
                    "regression": regression,
                    "advice": advice,
                    "softm": softm,
                    "alpha": alpha,
                    "beta": beta,
                }

                # Save a json file containing parameters and trees learned.
                model = [params, trees]
                Utils.save(".rfgb/models/" + target + ".json", model)

        models[target] = trees

    return models



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

"""
(docstring for utils)
"""

from random import sample
from math import exp

import codecs
import json
import string


class Data(object):
    """Object containing the relational data."""

    def __init__(
        self, regression=False, advice=False, softm=False, alpha=0.0, beta=0.0
    ):
        """
        An RFGB Data object, which serves as the structure for the positives,
        negatives, facts, and other parameters.

        adviceClauses: dictionary of advice clauses.
        facts: list of strings representing facts.
        pos: dictionary of positive examples.
        neg: dictionary of negative examples.
        examples: dictionary of examples for regression.
        examplesTrueValue: true value for use during regression.
        target: Target(s) to be learned or inferred.
        literals: Literals present in facts or their type specifications.
        variableType: Type of variable for facts and target.
        """

        self.regression = regression
        self.advice = advice
        self.adviceClauses = {}
        self.facts = []
        self.pos = {}
        self.neg = {}
        self.examples = {}
        self.examplesTrueValue = {}
        self.target = None
        self.literals = []
        self.literalTypes = {}
        self.variableType = {}

        self.softm = softm
        self.alpha = alpha
        self.beta = beta

    def setFacts(self, facts):
        """
        Mutate the facts in the data object.

        :param facts: List of strings representing the facts.
        :type facts: list.

        :returns: None
        """
        self.facts = facts

    def getFacts(self):
        """returns the facts in the data"""
        return self.facts

    def setPos(self, pos, target):
        """
        Set positive examples based on the contents of a list.
        """
        for example in pos:
            if example.split("(")[0] == target:
                # Set initial gradient to 0.5 for positives.
                self.pos[example] = 1 - Utils.sigmoid(-1.8)

    def setExamples(self, examples, target):
        """
        Set examples for regression.
        """
        for example in examples:
            # Get the predicate.
            predicate = example.split(" ")[0]
            # Get the true regression value.
            value = float(example.split(" ")[1])
            if predicate.split("(")[0] == target:
                # Store the true value in examplesTrueValue dictionary.
                self.examplesTrueValue[predicate] = value
                # Set the value, otherwise none.
                self.examples[predicate] = value

    def setNeg(self, neg, target):
        """
        Set negative examples based on the contents of a list.
        """
        for example in neg:
            if example.split("(")[0] == target:
                # Set initial gradient to -0.5 for negative examples.
                self.neg[example] = -Utils.sigmoid(-1.8)

    def setTarget(self, bk, target):
        """
        Sets self.target as a target string.
        Sets self.variableType

        :param bk: List of strings representing modes.
        :type bk: list.

        :param target: Target relation or attribute.
        :type target: str.

        :returns: None

        Example:

        .. code-block:: python

                        from rfgb.utils import Data

                        data = Data(regression=False)
                        background = ['friends(+person,-person)',
                                      'friends(-person,+person)',
                                      'smokes(+person)',
                                      'cancer(-person)']
                        target = 'cancer'

                        data.setTarget(background, target)

                        print(data.target)
                        # 'cancer(C)'
        """
        # targetTypes are the types of variables in the target predicate.
        targetTypes = [i[:-1].split("(")[1].split(",") for i in bk if target in i][0]
        targetTypes = list(map(Utils.removeModeSymbols, targetTypes))

        targetArity = len(targetTypes)
        targetVariables = sample(Utils.UniqueVariableCollection, targetArity)

        self.target = target + "("
        for variable in targetVariables:
            self.target += variable + ","
            self.variableType[variable] = targetTypes[targetVariables.index(variable)]
        self.target = self.target[:-1] + ")"

    def getTarget(self):
        """
        Returns the target.
        """
        return self.target

    def getExampleTrueValue(self, example):
        """
        Returns true regression value of an example for regression learning.
        """
        return self.examplesTrueValue[example]

    def getValue(self, example):
        """
        Returns the regression value for an example.

        Example:

        .. code-block:: python

                        from rfgb.utils import Utils
                        from rfgb.utils import Data

                        trainingData = Utils.readTrainingData('cancer',
                                            path='testDomain/ToyCancer/train/')

                        x = trainingData.getValue('cancer(earl)')
                        # x == -0.5, since earl doesn't have cancer.

                        y = trainingData.getValue('cancer(alice)')
                        # y == 0.5, since alice does have cancer
        """
        if self.regression:
            return self.examples[example]

        if example in self.pos:
            return self.pos[example]
        else:
            return self.neg[example]

    def setBackground(self, bk):
        """
        Obtains the literals and their type specifications. Types can be
        either variable or a list of constants.
        """

        bkWithoutTargets = [line for line in bk if "+" in line or "-" in line]

        # For every literal, obtain name and type specifications.
        for literalBk in bkWithoutTargets:
            literalName = literalBk.split("(")[0]
            literalTypeSpecification = literalBk[:-1].split("(")[1].split(",")
            self.literalTypes[literalName] = literalTypeSpecification
            self.literals.append([literalName, literalTypeSpecification])

    def getLiterals(self):
        """gets all the literals in the facts"""
        return self.literals

    def variance(self, examples):
        """
        Calculates the variance of the regression values from a subset of the
        data.
        """

        if not examples:
            return 0

        total = sum([self.getValue(example) for example in examples])
        numberOfExamples = len(examples)
        mean = total / float(numberOfExamples)
        sumOfSquaredError = sum(
            [(self.getValue(example) - mean) ** 2 for example in examples]
        )

        return sumOfSquaredError / float(numberOfExamples)  # return variance


class Utils(object):
    """
    Class of utilities used by rfgb, such as reading files, removing mode
    symbols, calculating Cartesian Products, etc.
    """

    # Attribute to store data (facts, positves, negatives)
    data = None
    UniqueVariableCollection = set(list(string.ascii_uppercase))

    @staticmethod
    def sigmoid(x):
        """
        :param x: Number to apply sigmoid to.
        :type x: int or float

        :returns: ``exp(x)/float(1+exp(x))``
        :rtype: float
        """
        return exp(x) / float(1 + exp(x))

    @staticmethod
    def removeModeSymbols(inputString):
        """
        Returns a string with the mode symbols (+,-,#) removed.

        Example:

        .. code-block:: python

                        from rfgb.utils import Utils

                        removeModeSymbols('#city')
                        # == 'city'

                        i = ['+drinks', '-drink', '-city']
                        o = list(map(removeModeSymbols, i))
                        # o == ['drinks', 'drink', 'city']
        """
        return inputString.replace("+", "").replace("-", "").replace("#", "")

    @staticmethod
    def addVariableTypes(literal):
        """
        As literals are encountered, update Utils.data.variableType with the
        type of the variables encountered.

        :param literal: A literal of the form smokes(W) or friends(A,B)
        :type literal: str.
        """

        # Get the name of the literal.
        literalName = literal.split("(")[0]

        # Get background info for the literal
        literalTypeSpecification = Utils.data.literalTypes[literalName]

        # Get the arguments
        literalArguments = literal[:-1].split("(")[1].split(",")

        # Get the number of arguments.
        for i in range(len(literalArguments)):
            if literalTypeSpecification[i][0] != "[":
                variable = literalArguments[i]
                if variable not in Utils.data.variableType.keys():
                    Utils.data.variableType[variable] = literalTypeSpecification[i][1:]

    @staticmethod
    def getleafValue(examples):
        """returns average of regression values for examples"""
        if not examples:
            return 0
        total = 0
        for example in examples:
            total += Utils.data.getValue(example)
        return total / float(len(examples))

    @staticmethod
    def save(location, saveItem):
        """
        Dumps json version of learnedDecisionTree to location.

        :param location: Name of the file to write.
        :type location: str.

        :returns: None.
        """
        with codecs.open(location, encoding="utf-8", mode="w") as f:
            json.dump(saveItem, f, indent=2)

    @staticmethod
    def load(location):
        """
        Loads json version of learnedDecisionTree from location.

        :param location: Name of the file to load.
        :type location: str.

        :returns: None.
        """
        with codecs.open(location, encoding="utf-8", mode="r") as f:
            return json.load(f)

    @staticmethod
    def readTrainingData(
        target,
        path="train/",
        regression=False,
        advice=False,
        softm=False,
        alpha=0.0,
        beta=0.0,
    ):
        """
        Reads the training data from files.

        :param target: The target predicate.
        :type target: str.

        :param path: Path to the training data.
        :type path: str.

        :param regression: Read from ``examples.txt`` instead of ``pos.txt``
                           and ``neg.txt``.
        :type regression: bool

        :param advice: Read advice from an advice file, which should be
                       contained in the same directory as the examples.
        :type advice: bool

        :default path: 'train/'
        :default regression: False
        :default advice: False

        :returns: A Data object representing the training data.
        :rtype: :py:class:`.utils.Data`
        """

        Utils.data = Data(
            regression=regression, advice=advice, softm=softm, alpha=alpha, beta=beta
        )
        # trainData = Data(regression=regression, advice=advice)
        # Utils.data.regression = regression
        # Utils.data.advice = advice

        if advice:
            with open(path + "advice.txt") as fp:  # read advice from train folder
                adviceFileLines = fp.read().splitlines()

                for line in adviceFileLines:
                    adviceClause = line.split(" ")[0]  # get advice clause

                    Utils.data.adviceClauses[adviceClause] = {}
                    # trainData.adviceClauses[adviceClause] = {}

                    preferredTargets = line.split(" ")[1][1:-1].split(",")
                    if preferredTargets[0]:
                        Utils.data.adviceClauses[adviceClause][
                            "preferred"
                        ] = preferredTargets
                        # trainData.adviceClauses[adviceClause]['preferred'] = preferredTargets
                    elif not preferredTargets[0]:
                        Utils.data.adviceClauses[adviceClause]["preferred"] = []
                        # trainData.adviceClauses[adviceClause]['preferred'] = []

                    nonPreferredTargets = line.split(" ")[2][1:-1].split(",")
                    if nonPreferredTargets[0]:
                        Utils.data.adviceClauses[adviceClause][
                            "nonPreferred"
                        ] = nonPreferredTargets
                        # trainData.adviceClauses[adviceClause]['nonPreferred'] = nonPreferredTargets
                    elif not nonPreferredTargets[0]:
                        Utils.data.adviceClauses[adviceClause]["nonPreferred"] = []
                        # trainData.adviceClauses[adviceClause]['nonPreferred'] = []

        with open(path + "facts.txt") as fac:
            Utils.data.setFacts(fac.read().splitlines())
            # trainData.setFacts(fac.read().splitlines())

        if regression:
            with open(path + "examples.txt") as exam:
                Utils.data.setExamples(exam.read().splitlines(), target)
                # trainData.setExamples(exam.read().splitlines(), target)
        else:
            with open(path + "pos.txt") as pos:
                Utils.data.setPos(pos.read().splitlines(), target)
                # trainData.setPos(pos.read().splitlines(), target)
            with open(path + "neg.txt") as neg:
                Utils.data.setNeg(neg.read().splitlines(), target)
                # trainData.setNeg(neg.read().splitlines(), target)

        with open(path + "bk.txt") as fp:
            bk = fp.read().splitlines()

            Utils.data.setBackground(bk)
            Utils.data.setTarget(bk, target)
            # trainData.setBackground(bk)
            # trainData.setTarget(bk, target, regression=regression)

        return Utils.data
        # return trainData

    @staticmethod
    def readTestData(target, path="test/", regression=False):
        """
        Reads the testing data from files.

        :param target: The target predicate.
        :type target: str.

        :param path: Path to the training data.
        :type path: str.

        :param regression: Read from ``examples.txt`` instead of ``pos.txt``
                           and ``neg.txt``.
        :type regression: bool

        :default path: 'train/'
        :default regression: False

        :returns: A Data object representing the training data.
        :rtype: :py:class:`.utils.Data`
        """

        testData = Data()
        testData.regression = regression

        with open(path + "facts.txt") as facts:
            testData.setFacts(facts.read().splitlines())

        if regression:
            with open(path + "examples.txt") as exam:
                examples = exam.read().splitlines()
                testData.setExamples(examples, target)
        else:
            # If we are not using regression, read from pos.txt and neg.txt
            with open(path + "pos.txt") as pos:
                testData.setPos(pos.read().splitlines(), target)
            with open(path + "neg.txt") as neg:
                testData.setNeg(neg.read().splitlines(), target)

        return testData

    @staticmethod
    def cartesianProduct(itemSets):
        """
        Returns the Cartesian Product of all sets contained in the item sets.
        """

        # Create new input where each element is in its own set.
        modifiedItemSets = []
        for itemSet in itemSets:
            modifiedItemSet = []
            for element in itemSet:
                modifiedItemSet.append([element])
            modifiedItemSets.append(modifiedItemSet)

        # Perform Cartesian Product of the first two sets.
        while len(modifiedItemSets) > 1:
            set1 = modifiedItemSets[0]
            set2 = modifiedItemSets[1]
            pairWiseProducts = []
            for item1 in set1:
                for item2 in set2:
                    # Cartesian Product performed here.
                    pairWiseProducts.append(item1 + item2)

            # Remove the first two sets.
            modifiedItemSets.remove(set1)
            modifiedItemSets.remove(set2)
            # Insert the Cartesian Product in their place and repeat.
            modifiedItemSets.insert(0, pairWiseProducts)

        # Return the final Cartesian Product Sets
        return modifiedItemSets[0]



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

"""
(docstring)
"""

from __future__ import print_function

from copy import deepcopy
from random import sample
import itertools
import re

from .utils import Utils

# Thanks to Chris Meyers for some of this code:
# http://www.openbookproject.net/py4fun/prolog/prolog1.html.


class Term(object):
    """
    Class for term in prolog proof.
    """

    # Expect "x(y,z...)"
    def __init__(self, s):
        if s[-1] != ")":
            raise (Exception("Term should end with ')': %s" % [s]))
        flds = s.split("(")
        if len(flds) != 2:
            raise (Exception("Term should be composed of two fields: %s" % [s]))
        self.args = flds[1][:-1].replace(" ", "").split(",")
        self.pred = flds[0]


class Rule(object):
    """
    Class for logic rules in prolog proof.
    """

    # Expect "term:-term;term;..."
    def __init__(self, s):
        flds = s.split(":-")
        self.head = Term(flds[0])
        self.goals = []
        if len(flds) == 2:
            flds = re.sub("\),", ");", flds[1]).split(";")
            for fld in flds:
                self.goals.append(Term(fld))


class Goal(object):
    """class for each goal in rule during prolog search"""

    def __init__(self, rule, parent=None, env={}):
        goalId = Prover.goalId
        goalId += 1
        self.id = goalId
        self.rule = rule
        self.parent = parent
        self.env = deepcopy(env)
        # Start search with 1st subgoal
        self.inx = 0


class Prover(object):
    """class for prolog style proof of query"""

    rules = []
    goalId = 100
    trace = 0

    @staticmethod
    def unify(srcTerm, srcEnv, destTerm, destEnv):
        """
        Unification method.
        """
        nargs = len(srcTerm.args)
        if nargs != len(destTerm.args):
            return 0
        if srcTerm.pred != destTerm.pred:
            return 0
        for i in range(nargs):
            srcArg = srcTerm.args[i]
            destArg = destTerm.args[i]
            if srcArg <= "Z":
                srcVal = srcEnv.get(srcArg)
            else:
                srcVal = srcArg
            if srcVal:
                # Constant or defined Variable in source
                if destArg <= "Z":
                    # Variable in destination
                    destVal = destEnv.get(destArg)
                    if not destVal:
                        # Unify
                        destEnv[destArg] = srcVal
                    elif destVal != srcVal:
                        # Won't unify
                        return 0
                elif destArg != srcVal:
                    # Won't unify
                    return 0
        return 1

    @staticmethod
    def search(term):
        """
        Method to perform prolog style query search.
        """
        goalId = Prover.goalId
        trace = Prover.trace
        rules = Prover.rules
        unify = Prover.unify
        goalId = 0
        returnValue = False
        if trace:
            print("search", term)

        # Anything- just get a rule object.
        goal = Goal(Rule("got(goal):-x(y)"))
        # Target is the single goal
        goal.rule.goals = [term]
        if trace:
            print("stack", goal)

        # Begin the search.
        stack = [goal]
        while stack:
            # Next goal to consider:
            c = stack.pop()
            if trace:
                print("  pop", c)

            # Is this one finished?
            if c.inx >= len(c.rule.goals):
                # Yes. Our original goal?
                if c.parent is None:
                    if c.env:
                        # Yes. tell user we
                        print(c.env)
                    else:
                        # have a solution
                        returnValue = True
                    continue

                # Otherwise, resume parent goal.
                parent = deepcopy(c.parent)

                unify(c.rule.head, c.env, parent.rule.goals[parent.inx], parent.env)

                # Advance to next goal in body.
                parent.inx = parent.inx + 1
                if trace:
                    print("stack", parent)
                # Let it wait its turn.
                stack.append(parent)
                continue

            # No. more to do with this goal.
            # What we want to solve:
            term = c.rule.goals[c.inx]
            for rule in rules:
                # Walk down the rule database.
                if rule.head.pred != term.pred:
                    continue
                if len(rule.head.args) != len(term.args):
                    continue
                # A possible subgoal.
                child = Goal(rule, c)
                ans = unify(term, c.env, rule.head, child.env)
                if ans:
                    # if unifies, stack it up
                    if trace:
                        print("stack", child)
                    stack.append(child)
        return returnValue

    @staticmethod
    def prove(data, example, clause):
        """
        Proves if example satisfies clause given the data.
        Returns True if it satisfies, else return False.

        Prover.rules: contains all of the rules.
        Prover.trace: If this is 1, displays the proof tree.
        Prover.goalID: stores the goal ID.
        """
        Prover.rules = []
        Prover.trace = 0
        Prover.goalId = 100
        Prover.rules += [Rule(fact) for fact in data.getFacts()]
        Prover.rules += [Rule(clause)]

        # Proves query prolog-style:
        proofOutcome = Prover.search(Term(example))
        return proofOutcome


class Logic(object):
    """
    Class for logic operations.
    """

    @staticmethod
    def constantsPresentInLiteral(literalTypeSpecification):
        """
        Returns true if constants present in type specification.
        """
        # Check if there is a single non-variable.
        for item in literalTypeSpecification:
            if item[0] == "[":
                return True
        return False

    @staticmethod
    def getVariables(literal):
        """
        Returns variables in the literal.
        """
        # Get variables and constants in body literal.
        variablesAndConstants = literal[:-1].split("(")[1].split(",")

        # Get only the variables.
        variables = []
        for item in variablesAndConstants:
            if item in Utils.UniqueVariableCollection:
                variables.append(item)

        return variables

    @staticmethod
    def generateTests(literalName, literalTypeSpecification, clause):
        """
        Generates tests for literal according to modes and types.
        """

        target = clause.split(":-")[0]
        body = clause.split(":-")[1]
        targetVariables = target[:-1].split("(")[1].split(",")

        # Initialize a list of body variables.
        bodyVariables = []
        if body:
            # Get clause body literals
            bodyLiterals = [literal for literal in body.split(";") if literal]
            for literal in bodyLiterals:
                bodyVariables += Logic.getVariables(literal)

        clauseVariables = set(targetVariables + bodyVariables)
        lengthOfSpecification = len(literalTypeSpecification)

        testSpecification = []
        for i in range(lengthOfSpecification):
            # Check if data type is variable or constant.
            variable = False
            if literalTypeSpecification[i][0] != "[":
                variable = True
            # If the data type is variable.
            if variable:
                # Get mode (+ or -)
                mode = literalTypeSpecification[i][0]
                # Get the variable type.
                variableType = literalTypeSpecification[i][1:]

                # Variable must be an already existing variable in the clause
                # of the same type if it exists.
                if mode == "+":
                    # Get all clause variables of same type:

                    variableOfSameTypeInClause = []
                    for var in clauseVariables:
                        if Utils.data.variableType[var] == variableType:
                            variableOfSameTypeInClause.append(var)

                    if variableOfSameTypeInClause:
                        # If variables of same type exist in clause
                        testSpecification.append(variableOfSameTypeInClause)
                    else:
                        newVar = None
                        while True:
                            newVar = sample(Utils.UniqueVariableCollection, 1)
                            if newVar[0] not in clauseVariables:
                                break
                        testSpecification.append([newVar[0]])

                # Use new variable.
                if mode == "-":
                    newVar = None
                    while True:
                        newVar = sample(Utils.UniqueVariableCollection, 1)
                        if newVar[0] not in clauseVariables:
                            break
                    testSpecification.append([newVar[0]])

            # If data type is constant:
            else:
                listToAppend = literalTypeSpecification[i][1:-1].split(";")
                testSpecification.append(listToAppend)

        testVariablesAndConstants = Utils.cartesianProduct(testSpecification)
        literalCandidates = []

        # Form predicates and return all the test candidates for this literal
        for item in testVariablesAndConstants:
            literalCandidate = literalName + "(" + ",".join(item) + ")"
            literalCandidates.append(literalCandidate)
        return literalCandidates



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

from __future__ import print_function
from __future__ import absolute_import

from ..boosting import performInference
from ..utils import Utils


def infer(target, trees, path="", regression=False):
    """
    .. versionadded:: 0.3.0

    Perform inference on data with a set of trees.

    .. note:: This currently requires that test data is stored as files
              on disk.

    :param trees: Trees to perform inference with.
    :type trees: list of str.

    :param path: Path to the location test data is stored.
    :type path: str.

    :param regression: Infer with a regression model instead of classification.
    :type regression: bool.

    :default regression: False

    :returns: Tuple of results. In classification these results will be a tuple
              of positive and negative examples. In regression this will be
              the examples.
    :rtype: tuple
    """

    # Read the testing data.
    testData = Utils.readTestData(target, path=path, regression=regression)

    # Get the probability of the test examples.
    performInference(testData, trees)

    if regression:
        return testData.examples
    else:
        return (testData.pos, testData.neg)



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

"""
Core methods for performing learning and inference, such as computing
gradients, updating gradients, and performing inference.

Documentation
-------------
"""

from __future__ import division

from .utils import Utils
from .logic import Prover

from math import log
from math import exp
from copy import deepcopy

_log_prior = -1.8


def computeAdviceGradient(example):
    """
    Proves each clause (:meth:`.Prover.prove`) and computes the advice gradient
    as ``NumberTrue - NumberFalse``.

    :param example:
    :type example:
    """

    nt, nf = 0, 0
    target = Utils.data.target.split("(")[0]
    for clause in Utils.data.adviceClauses:
        if Prover.prove(Utils.data, example, clause):
            if target in Utils.data.adviceClauses[clause]["preferred"]:
                nt += 1
            if target in Utils.data.adviceClauses[clause]["nonPreferred"]:
                nf += 1
    return nt - nf


def computeSumOfGradients(example, trees, data):
    """
    Computes new gradients for an example.

    :param example:
    :type example:

    :param trees:
    :type trees:

    :param data:
    :type data:
    """

    sumOfGradients = 0

    # Add leaf values satisfied by example in each tree.
    for tree in trees:
        gradient = inferTreeValue(tree, example, data)
        sumOfGradients += gradient
    return sumOfGradients


def inferTreeValue(clauses, query, data):
    """
    Returns the probability of `query` given data and the clauses learned.

    :param clauses:
    :type clauses:

    :param query:
    :type query:

    :param data:
    :type data:
    """
    for clause in clauses:

        clauseCopy = deepcopy(clause)
        clauseValue = float(clauseCopy.split(" ")[1])
        clauseRule = clauseCopy.split(" ")[0].replace(";", ",")

        if not clauseRule.split(":-")[1]:
            return clauseValue
        if Prover.prove(data, query, clauseRule):
            # Check if query satisifes clause
            return clauseValue


def performInference(testData, trees):
    """
    Computes the probabilities for test examples.

    :param testData: Data for testing.
    :type testData: :py:class:`.utils.Data` object.

    :param trees: List of strings representing learned decision trees.
    :type trees: list.

    Example:

    .. code-block:: python

                    from rfgb.boosting import performInference

    """

    logPrior = _log_prior
    if not testData.regression:

        # Initialize log odds of assumed prior probability for example.

        for example in testData.pos:

            # Compute sum of gradients
            sumOfGradients = computeSumOfGradients(example, trees, testData)

            # Calculate probability as sigmoid(log odds)
            testData.pos[example] = Utils.sigmoid(logPrior + sumOfGradients)

        for example in testData.neg:

            # Compute sum of gradients
            sumOfGradients = computeSumOfGradients(example, trees, testData)

            # Calculate probability as sigmoid(log odds)
            testData.neg[example] = Utils.sigmoid(logPrior + sumOfGradients)

    elif testData.regression:
        logPrior = 0.0
        for example in testData.examples:
            sumOfGradients = computeSumOfGradients(example, trees, testData)
            testData.examples[example] = sumOfGradients


def updateGradients(data, trees, loss="LS", delta=None):
    """
    Update gradients of the data.

    :param data: Training or testing data (with parameters).
    :type data: :py:class:`.utils.Data` object.

    :param trees: List of strings representing trees.
    :type trees: list.

    :param loss: Loss function for regression (currently implemented:
                 'LS', 'LAD', 'Huber').
    :type loss: str.

    :param delta: Delta value for Huber loss.
    :type delta: float

    Example:

    .. code-block:: python

                    from rfgb.boosting import updateGradients

    """

    if data.regression:
        # If this is regression data, compute gradient as y - y_hat

        for example in data.examples:
            sumOfGradients = computeSumOfGradients(example, trees, data)
            trueValue = data.getExampleTrueValue(example)

            if loss == "LS":
                # Least Squares
                data.examples[example] = trueValue - sumOfGradients

            elif loss == "LAD":
                # Least Absolute Deviation
                updatedGradient = 0
                gradient = trueValue - sumOfGradients
                if gradient:
                    updatedGradient = gradient / float(abs(gradient))
                data.examples[example] = updatedGradient

            elif loss == "Huber":
                # Huber Loss
                gradient = trueValue - sumOfGradients
                updatedGradient = 0
                if gradient:
                    if gradient > float(delta):
                        updatedGradient = gradient / float(abs(gradient))
                    elif gradient <= float(delta):
                        updatedGradient = gradient
                data.examples[example] = updatedGradient

    else:
        # If this is classification data, compute 1 - P for each positive.
        """

        logPrior = __logPrior__

        # P = sigmoid(sum of gradients) given by each tree learned so far.

        for example in data.pos:
            # For each positive example compute 1 - P
            sumOfGradients = computeSumOfGradients(example, trees, data)
            prob = Utils.sigmoid(logPrior + sumOfGradients)
            updatedGradient = 1 - prob
            if data.advice:
                adviceGradient = computeAdviceGradient(example)
                updatedGradient += adviceGradient
            data.pos[example] = updatedGradient

        for example in data.neg:
            # For each negative example compute 0 - P
            sumOfGradients = computeSumOfGradients(example, trees, data)
            prob = Utils.sigmoid(logPrior + sumOfGradients)
            updatedGradient = 0 - prob
            if data.advice:
                adviceGradient = computeAdviceGradient(example)
                updatedGradient += adviceGradient
            data.neg[example] = updatedGradient

        """

        logPrior = _log_prior

        if data.softm:

            for example in data.pos:

                sumOfGradients = computeSumOfGradients(example, trees, data)
                prob = Utils.sigmoid(logPrior + sumOfGradients)
                updatedGradient = 1 - prob / (prob + (1 - prob) * exp(data.alpha))
                data.pos[example] = updatedGradient

            for example in data.neg:

                sumOfGradients = computeSumOfGradients(example, trees, data)
                prob = Utils.sigmoid(logPrior + sumOfGradients)
                updatedGradient = 1 - prob / (prob + (1 - prob) * exp(-data.beta))
                data.neg[example] = updatedGradient

        else:

            for example in data.pos:

                sumOfGradients = computeSumOfGradients(example, trees, data)
                prob = Utils.sigmoid(logPrior + sumOfGradients)
                updatedGradient = 1 - prob

                if data.advice:
                    updatedGradient += computeAdviceGradient(example)

                data.pos[example] = updatedGradient

            for example in data.neg:

                sumOfGradients = computeSumOfGradients(example, trees, data)
                prob = Utils.sigmoid(logPrior + sumOfGradients)
                updatedGradient = 0 - prob

                if data.advice:
                    updatedGradient += computeAdviceGradient(example)

                data.neg[example] = updatedGradient



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

"""
rfgb
====

Provides
    1. The ability to perform Relational Functional Gradient Boosting (RFGB)
       in Python.

Documentation
-------------

The vast majority of documentation is currently provided in-line with the
code, mostly consisting of high-level overviews and some usage examples.

As this grows in complexity, documentation will likely be found in either
the doc strings for individual functions (i.e. rfgb.__doc__), or online.
"""

from .utils import Utils
from .tree import node
from .boosting import updateGradients
from .boosting import performInference

from ._metadata import (
    __author__,
    __maintainer__,
    __email__,
    __copyright__,
    __license__,
    __version__,
    __status__,
)



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

"""
.. versionadded:: 0.3.0

Learn and infer with relational dependency networks.

.. code-block:: python

    # Example script for performing learning and inference.

    from rfgb import rdn

    # rdn.learn requires a list of targets as strings.
    trees = rdn.learn(['cancer'], path='testDomains/ToyCancer/train/')

    # rdn.learn returns a dictionary mapping targets to trees.
    cancer_trees = trees['cancer']

    # rdn.infer classification returns a tuple of pos and neg.
    results = rdn.infer('cancer', cancer_trees, path='testDomains/ToyCancer/test/')

    # ({'cancer(xena)': 0.34460796550872186,
    #   'cancer(yoda)': 0.34460796550872186,
    #   'cancer(zod)': 0.34460796550872186},
    #  {'cancer(watson)': 0.34460796550872186,
    #   'cancer(voldemort)': 0.34460796550872186})
"""

from .learn import learn
from .infer import infer



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>


"""
Unit tests for Utils.py

Tests should be ordered in the same manner as they are defined in Utils.py
"""

from __future__ import absolute_import
from ...utils import Data
from ...utils import Utils
import unittest


class UtilsTest(unittest.TestCase):
    def test_sigmoid(self):
        """
        tests: sigmoid function.
        """
        self.assertEqual(Utils.sigmoid(-5), 0.006692850924284856)
        self.assertEqual(Utils.sigmoid(-0.001), 0.49975000002083336)
        self.assertEqual(Utils.sigmoid(0), 0.5)
        self.assertEqual(Utils.sigmoid(0.0001), 0.5000249999999792)
        self.assertEqual(Utils.sigmoid(0.01), 0.5024999791668749)
        self.assertEqual(Utils.sigmoid(0.5), 0.6224593312018546)
        self.assertEqual(Utils.sigmoid(1), 0.7310585786300049)
        self.assertEqual(Utils.sigmoid(5), 0.9933071490757152)

    def test_removeModeSymbols(self):
        """
        tests: removeModeSymbols function.
        """
        self.assertEqual(Utils.removeModeSymbols("+drinks"), "drinks")
        self.assertEqual(Utils.removeModeSymbols("-drinks"), "drinks")
        self.assertEqual(Utils.removeModeSymbols("#drinks"), "drinks")

    def test_setTarget_1(self):
        """
        tests: Data.setTarget (attribute case).
        """

        # Some setup that needs to take place.

        target = "cancer"
        background = [
            "friends(+person,-person)",
            "friends(-person,+person)",
            "smokes(+person)",
            "cancer(person)",
        ]

        # setTarget has some random behavior, but always returns the target
        # bound to a letter.
        for _ in range(100):
            # Set the target (introducing a random variable).
            data = Data(regression=False)
            data.setBackground(background)
            data.setTarget(background, target)

            # Assert the presence of the target.
            self.assertTrue(target in data.target)

            # Get the random variable.
            instance = data.target.split("(")[1][0]

            # Assert that the random variable belongs to the
            # UniqueVariableCollection.
            self.assertTrue(instance in Utils.UniqueVariableCollection)

    def test_setTarget_2(self):
        """
        tests: Data.setTarget (relation case).
        """

        # Some setup that needs to take place.
        target = "friends"
        background = [
            "friends(+person,-person)",
            "friends(-person,+person)",
            "smokes(+person)",
            "cancer(person)",
        ]

        for _ in range(100):
            # Set the target (introducing two random variables).
            data = Data(regression=False)
            data.setBackground(background)
            data.setTarget(background, target)

            # Assert the presence of the target.
            self.assertTrue(target in data.target)

            # Get the two random variables.
            instance = data.target.split("(")[1].split(")")[0].split(",")

            # Assert arity of friendship is 2, and that variables belong to
            # UniqueVariableCollection.
            self.assertEqual(len(instance), 2)
            self.assertTrue(instance[0] in Utils.UniqueVariableCollection)
            self.assertTrue(instance[1] in Utils.UniqueVariableCollection)

    def test_setTarget_3(self):
        """
        tests: Data.setTarget (3-arity case).
        """

        target = "drinks"
        background = ["drinks(+person,-drink,-city)"]

        for _ in range(100):
            # Set the target (introducing three random variables).
            data = Data(regression=False)
            data.setBackground(background)
            data.setPos(
                ["drinks(Jeremy,Wine,Paris)", "drinks(Ulrich,Beer,Frankfurt)"], target
            )
            data.setTarget(background, target)

            # Assert the presence of the target.
            self.assertTrue(target in data.target)

            # Get the three random variables.
            instance = data.target.split("(")[1].split(")")[0].split(",")

            # Assert arity of drinks is 3, variables are not duplicated, and
            # belong to UniqueVariableCollection.
            self.assertEqual(len(instance), 3)
            self.assertTrue(instance[0] in Utils.UniqueVariableCollection)
            self.assertTrue(instance[1] in Utils.UniqueVariableCollection)
            self.assertTrue(instance[2] in Utils.UniqueVariableCollection)

    def test_read_test_data_toy_cancer(self):
        """
        tests: readTestData on the ToyCancer dataset.
        """
        sampleData = Utils.readTestData("cancer", path="testDomains/ToyCancer/test/")

        # Verify that regression and advice were not specified.
        self.assertEqual(sampleData.regression, False)
        self.assertEqual(sampleData.advice, False)

        # sampleData.facts should return a list of strings.
        self.assertEqual(
            sampleData.facts,
            [
                "friends(zod,xena)",
                "friends(xena,watson)",
                "friends(watson,voldemort)",
                "friends(voldemort,yoda)",
                "friends(yoda,zod)",
                "friends(xena,zod)",
                "friends(watson,xena)",
                "friends(voldemort,watson)",
                "friends(yoda,voldemort)",
                "friends(zod,yoda)",
                "smokes(zod)",
                "smokes(xena)",
                "smokes(yoda)",
            ],
        )

        # sampleData.pos should return a dictionary with the
        # initial gradients set to 0.5
        self.assertEqual(
            sampleData.pos,
            {
                "cancer(zod)": 0.8581489350995122,
                "cancer(xena)": 0.8581489350995122,
                "cancer(yoda)": 0.8581489350995122,
            },
        )
        # sampleData.neg should return a dictionary with the
        # initial gradients set to -0.5
        self.assertEqual(
            sampleData.neg,
            {
                "cancer(voldemort)": -0.1418510649004878,
                "cancer(watson)": -0.1418510649004878,
            },
        )

        # sampleData.literals should be an empty dictionary.
        self.assertEqual(sampleData.literals, [])

        # Test some of the Data object functions, they should not
        # mutate contents.
        self.assertEqual(sampleData.facts, sampleData.getFacts())
        self.assertEqual(sampleData.literals, sampleData.getLiterals())

    def test_read_test_data_boston_housing(self):
        """
        tests: readTestData on the BostonHousing dataset.
        """
        sampleData = Utils.readTestData(
            "medv", path="testDomains/BostonHousing/test/", regression=True
        )

        # Verify that regression was specified but advice was not.
        self.assertEqual(sampleData.regression, True)
        self.assertEqual(sampleData.advice, False)

        # sampleData.facts should return a list of strings
        with open("testDomains/BostonHousing/test/facts.txt") as f:
            bostonHousingFacts = f.read().splitlines()
        self.assertEqual(sampleData.facts, bostonHousingFacts)

        # sampleData.examples should return a dictionary where the first
        # item is the predicate and the second item is the regression value.
        with open("testDomains/BostonHousing/test/examples.txt") as f:
            bostonHousingExamples = f.read().splitlines()

        for e in bostonHousingExamples:
            ePred, eVal = e.split()
            self.assertTrue(ePred in sampleData.examples)
            self.assertTrue(float(eVal) in list(sampleData.examples.values()))

        self.assertEqual(
            sampleData.examples,
            {
                "medv(id348)": 23.1,
                "medv(id369)": 50.0,
                "medv(id344)": 23.9,
                "medv(id152)": 19.6,
                "medv(id120)": 19.3,
                "medv(id53)": 25.0,
                "medv(id422)": 14.2,
                "medv(id115)": 18.5,
                "medv(id235)": 29.0,
                "medv(id448)": 12.6,
                "medv(id3)": 34.7,
                "medv(id211)": 21.7,
                "medv(id334)": 22.2,
                "medv(id313)": 19.4,
                "medv(id13)": 21.7,
                "medv(id42)": 26.6,
                "medv(id439)": 84.0,
                "medv(id45)": 21.2,
                "medv(id183)": 37.9,
                "medv(id212)": 19.3,
                "medv(id374)": 13.8,
                "medv(id238)": 31.5,
                "medv(id294)": 23.9,
                "medv(id7)": 22.9,
                "medv(id62)": 16.0,
                "medv(id153)": 15.3,
                "medv(id345)": 31.2,
                "medv(id163)": 50.0,
                "medv(id341)": 18.7,
                "medv(id50)": 19.4,
                "medv(id208)": 22.5,
                "medv(id167)": 50.0,
                "medv(id70)": 20.9,
                "medv(id143)": 13.4,
                "medv(id217)": 23.3,
                "medv(id29)": 18.4,
                "medv(id181)": 39.8,
                "medv(id342)": 32.7,
                "medv(id59)": 23.3,
                "medv(id27)": 16.6,
                "medv(id188)": 32.0,
                "medv(id435)": 11.7,
                "medv(id173)": 23.1,
                "medv(id10)": 18.9,
            },
        )



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

"""
_metadata.py
============

Metadata for rfgb package
"""

__author__ = "Alexander L. Hayes"
__maintainer__ = "Alexander L. Hayes"
__email__ = "hayesall@iu.edu"
__copyright__ = "Copyright © 2017-2019 rfgb Contributors"
__license__ = "GNU General Public License v3.0 (GPLv3)"
__version__ = "0.3.0"
__status__ = "Prototype"



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

"""
rfgb
----

(docstring for main function)
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from .boosting import updateGradients
from .boosting import performInference
from .tree import node
from .utils import Utils
from ._metadata import __version__

# Imports for subpackages
from . import cmd
from . import rdn

import argparse
import os
import sys

# Create an argument parser for interpretting user inputs.
PARSER = argparse.ArgumentParser(
    prog="rfgb",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=(
        "rfgb: Relational Functional Gradient Boosting is a gradient-boosting\n"
        "approach to learning statistical relational models.\n\n"
        "Start a working area (see also: rfgb help tutorial)\n"
        "   init    Initialize an empty .rfgb directory at current location."
    ),
    epilog=(
        "Copyright © 2017-2019 rfgb Contributors. Distributed under the terms\n"
        "of the GNU GPL version 3 or later. <http://gnu.org/licenses/gpl.html>\n"
        "This is free software: you are free to change and redistribute it.\n"
        "There is NO WARRANTY, to the extent permitted by law.\n"
    ),
)

# Create subparsers for learning different types of models, their respective
# arguments may be set individually (for example, if advice is implemented
# for RDNs but has not been worked out for MLNs).

# This is different from BoostSRL's codebase, where the default is RDN and
# MLNs are learned by supplying a -mln flag. This change is to make things
# more-easily extended in the future.
SUBPARSERS = PARSER.add_subparsers(
    title="rfgb Subcommands", help="$ rfgb --help", dest="_rfgb"
)

INIT_PARSER = SUBPARSERS.add_parser(
    "init",
    description=(
        "Initialize an rfgb package in the current directory for saving,\n"
        "loading, and managing models.",
    ),
    help="Initialize a .rfgb directory.",
)
HELP_PARSER = SUBPARSERS.add_parser(
    "help", description="Commandline references for common actions."
)

LEARN_PARSER = SUBPARSERS.add_parser(
    "learn", description="Specify the model to learn.", help="Learn various SRL models."
)

INFER_PARSER = SUBPARSERS.add_parser(
    "infer",
    description="Make inferences using the working model.",
    help="Infer with various SRL models.",
)

# Sub-commands specific to learning different models.
LEARN_SUBPARSER = LEARN_PARSER.add_subparsers(
    title="RFGB Learn",
    description="Learning statistical relational models.",
    help="$ rfgb learn --help",
    dest="_learn",
)
RDN_PARSER = LEARN_SUBPARSER.add_parser(
    "rdn",
    description="Relational Dependency Networks",
    help="Relational Dependency Networks",
)
MLN_PARSER = LEARN_SUBPARSER.add_parser(
    "mln", description="Markov Logic Networks", help="Markov Logic Networks"
)
SPN_PARSER = LEARN_SUBPARSER.add_parser(
    "spn", description="Sum-Product Networks", help="Sum-Product Networks"
)
RRBM_PARSER = LEARN_SUBPARSER.add_parser(
    "rrbm",
    description="Relational Restricted Boltzmann Machines",
    help="Relational Restricted Boltzmann Machines",
)

# infer-specific arguments
INFER_PARSER.add_argument(
    "-target",
    "--target",
    type=str,
    default=None,
    action="append",
    help="Target predicate(s) to infer.",
)

# init-specific arguments
INFER_PARSER.add_argument(
    "-test", "--test", type=str, default="test/", help="Set the testing directory."
)
INIT_PARSER.add_argument("-q", "--quiet", help="Quiet output.", action="store_true")

# RDN-specific arguments.
RDN_PARSER.add_argument(
    "-advice",
    "--advice",
    help=(
        "Trigger learning with expert advice. Currently reads from an "
        "advice.txt file stored in the same location as the data."
    ),
    action="store_true",
)
RDN_PARSER.add_argument(
    "-reg",
    "--regression",
    help="Learn a regression model instead of a classification model.",
    action="store_true",
)
RDN_PARSER.add_argument(
    "-softm",
    "--softm",
    help=(
        "Softmax boosting. Set the false positive weight (alpha)"
        "and false negative weight (beta) along with this parameter."
    ),
    default=False,
    action="store_true",
)
RDN_PARSER.add_argument(
    "-alpha",
    "--alpha",
    help="Set the alpha value for use with softm.",
    type=float,
    default=0.0,
)
RDN_PARSER.add_argument(
    "-beta",
    "--beta",
    help="Set the beta value for use with softm.",
    type=float,
    default=0.0,
)
RDN_PARSER.add_argument(
    "-trees",
    "--trees",
    type=int,
    default=10,
    help="Specify the number of boosted regression trees to learn (Default: 10)",
)
RDN_PARSER.add_argument(
    "-target",
    "--target",
    type=str,
    default=None,
    action="append",
    help="Target predicate(s) to learn.",
)
RDN_PARSER.add_argument(
    "-train", "--train", type=str, default="train/", help="Set the training directory."
)

# Get the arguments
PARAMETERS = PARSER.parse_args()

if PARAMETERS._rfgb == "init":
    # Initialize an empty rfgb repository for loading and saving models.
    cmd.init(quiet=PARAMETERS.quiet)

elif PARAMETERS._rfgb == "help":
    print("Help information.")

elif PARAMETERS._rfgb == "learn":

    if PARAMETERS._learn == "rdn":
        print("Learning RDN.")

        if not PARAMETERS.target:
            raise (ValueError("'target' must be provided."))

        TREES = rdn.learn(
            PARAMETERS.target,
            path=PARAMETERS.train,
            regression=PARAMETERS.regression,
            advice=PARAMETERS.advice,
            softm=PARAMETERS.softm,
            alpha=PARAMETERS.alpha,
            beta=PARAMETERS.beta,
        )

    elif PARAMETERS._learn == "mln":
        print("Learning MLN (TODO)")
        exit(1)

    elif PARAMETERS._learn == "spn":
        print("Learning SPN (TODO)")
        exit(1)

    elif PARAMETERS._learn == "rrbm":
        print("Learning RRBM (TODO)")
        exit(1)

elif PARAMETERS._rfgb == "infer":

    if PARAMETERS.target:
        TARGETS = PARAMETERS.target
    else:
        # Collect each target from the files in the .rfgb/models/ directory.
        TARGETS = list(
            map(lambda s: s.replace(".json", ""), os.listdir(".rfgb/models/"))
        )

    for target in TARGETS:

        model = Utils.load(".rfgb/models/" + target + ".json")
        settings, trees = model[0], model[1]

        results = rdn.infer(
            target, trees, path=PARAMETERS.test, regression=settings["regression"]
        )

        # Print results for easy viewing.
        print(results)

    exit(0)

else:
    print("Reached end of program without performing any action.")
    exit(1)
exit(0)



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

from __future__ import print_function
from __future__ import absolute_import
from ...tree import node
import unittest


class TreeTest(unittest.TestCase):
    def test_setNodeDepth(self):

        node.setMaxDepth(1)
        self.assertEqual(node.maxDepth, 1)

        node.setMaxDepth(2)
        self.assertEqual(node.maxDepth, 2)

        node.setMaxDepth(3)
        self.assertEqual(node.maxDepth, 3)

        node.setMaxDepth(100)
        self.assertEqual(node.maxDepth, 100)



"""
Setup file for rfgb
"""

from setuptools import setup
from setuptools import find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Get __version__ from _metadata.py
with open(path.join("rfgb", "_metadata.py")) as f:
    exec(f.read())

setup(
    name='rfgb',
    version=__version__,
    license=__license__,

    description='Relational Functional Gradient Boosting in Python.',
    long_description=long_description,

    author='Alexander L. Hayes',
    author_email='hayesall@iu.edu',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    keywords='relational-learning',
    project_urls={
        'Source Code': 'https://github.com/hayesall/rfgb',
        'Bug Tracker': 'https://github.com/hayesall/rfgb/issues',
        'Documentation': 'https://rfgb.readthedocs.io/en/latest/',
    },

    entry_points={
        'console_scripts': ['rfgb=rfgb.__main__']
    },

    packages=find_packages(exclude=['tests'])
)



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

from rfgbtests import *
import unittest

if __name__ == "__main__":
    """
    Main testing module for ``rfgb``, must be ran from the base of the
    project repository.

    For example:

    .. code-block:: bash

                    python rfgb/tests/tests.py

    Verbosity may be explicitly set by passing an integer with the ``-v``
    flag. The value will be passed into unittest.TextTestRunner, so integers
    higher than 1 will lead to more verbose outputs.

    .. code-block:: bash

                    python rfgb/tests/tests.py -v 2

    Individual modules may be tested with unittest via the command line.

    .. code-block:: bash

                    python -m unittest rfgb/tests/rfgbtests/test_Utils.py
                    .......
                    --------------------------------------------------
                    Ran 7 tests in 0.005s

                    OK
    """

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verbose", default=1, type=int)
    args = parser.parse_args()

    testsuite = unittest.TestLoader().discover(".")
    runner = unittest.TextTestRunner(verbosity=args.verbose)

    """
    @hayesall:

    I have not found a more elegant way to handle this problem, but
    to summarize: coverage and Travis-ci only seem to recognize errors
    when an exception is raised. TextTestRunner.run does not appear to
    raise exceptions on failures, so instead this must be done explicitly.
    """

    results = runner.run(testsuite)
    if results.failures or results.errors:
        raise (Exception("Encountered errors during runner.run"))



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>

"""
Test the init function twice. Once to create the directory, and
again to check whether an exception is raised properly if another
one is created.
"""

from __future__ import print_function
from __future__ import absolute_import

from ...cmd import init
import os
import sys
import unittest


def reset_directories():
    if os.path.exists(".rfgb/") or os.path.exists(".rfgb/models/"):
        os.rmdir(".rfgb/models/")
        os.rmdir(".rfgb/")


class InitTest(unittest.TestCase):
    """
    Tests for rfgb.cmd.init

    1. Make sure .rfgb directory does not currently exist.
    2. Run tests with both default parameter, and quiet parameter.
    """

    def test_init_default_run(self):
        reset_directories()

        # First run should create directories.
        init()
        self.assertTrue(os.path.exists(".rfgb/"))
        self.assertTrue(os.path.exists(".rfgb/models/"))

        # Second should exit and print message to console.
        with self.assertRaises(SystemExit) as ExitCode:
            init()
        self.assertEqual(ExitCode.exception.code, 1)

    def test_init_quiet_run(self):
        reset_directories()

        # First run should create directories.
        init(quiet=True)
        self.assertTrue(os.path.exists(".rfgb/"))
        self.assertTrue(os.path.exists(".rfgb/models/"))

        # Second should exit and print message to console.
        with self.assertRaises(SystemExit) as ExitCode:
            init(quiet=True)
        self.assertEqual(ExitCode.exception.code, 1)


# Remove directories at end of test.
reset_directories()



# -*- coding: utf-8 -*-

# Copyright © 2017-2019 rfgb Contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (at the base of this repository). If not,
# see <http://www.gnu.org/licenses/>


from __future__ import print_function
from __future__ import absolute_import
from ...logic import Term
from ...logic import Rule
from ...logic import Logic

import unittest


class TermTest(unittest.TestCase):
    """
    Tests for rfgb.logic.Term
    """

    def test_initialize_term_0(self):
        term = Term("()")
        self.assertEqual(term.pred, "")
        self.assertEqual(term.args, [""])

    def test_initialize_term_1(self):
        term = Term("person(batflyer)")
        self.assertEqual(term.pred, "person")
        self.assertEqual(term.args, ["batflyer"])

    def test_initialize_term_2(self):
        term = Term("drinks(person,beer)")
        self.assertEqual(term.pred, "drinks")
        self.assertEqual(term.args, ["person", "beer"])

    def test_initialize_term_3(self):
        term = Term("drinks(person,beer,pub)")
        self.assertEqual(term.pred, "drinks")
        self.assertEqual(term.args, ["person", "beer", "pub"])

    def test_initialize_term_4(self):
        term = Term("drinks(person, beer, pub)")
        self.assertEqual(term.pred, "drinks")
        self.assertEqual(term.args, ["person", "beer", "pub"])

    def test_initialize_bad_term_1(self):
        # Exception should be raised on account of the last item in the string
        # not being a ')'
        with self.assertRaises(Exception):
            term = Term(")(")

    def test_initialize_bad_term_2(self):
        # Exception should be raised due to the syntax error at the end.
        with self.assertRaises(Exception):
            term = Term("hello(world).")

    def test_initialize_bad_term_3(self):
        # Similar reason to previous.
        with self.assertRaises(Exception):
            term = Term("drinks(person, pub).")

    def test_initialize_bad_term_4(self):
        # Terms should be composed of two fields split on a '('
        with self.assertRaises(Exception):
            term = Term("x)")


class RuleTest(unittest.TestCase):
    """
    Tests for rfgb.logic.Rule
    """

    def test_initialize_rule_0(self):
        rule = Rule("smokes(x):-cancer(x)")
        self.assertEqual(rule.head.pred, "smokes")
        self.assertEqual(rule.head.args, ["x"])
        self.assertEqual(rule.goals[0].pred, "cancer")
        self.assertEqual(rule.goals[0].args, ["x"])

    def test_initialize_rule_1(self):
        rule = Rule("advises(x,y):-paper(x,y)")
        self.assertEqual(rule.head.pred, "advises")
        self.assertEqual(rule.head.args, ["x", "y"])
        self.assertEqual(rule.goals[0].pred, "paper")
        self.assertEqual(rule.goals[0].args, ["x", "y"])

    def test_initialize_rule_2(self):
        rule = Rule("advises(x,y):-paper(x,y);student(y)")
        self.assertEqual(rule.head.pred, "advises")
        self.assertEqual(rule.head.args, ["x", "y"])
        self.assertEqual(rule.goals[0].pred, "paper")
        self.assertEqual(rule.goals[0].args, ["x", "y"])
        self.assertEqual(rule.goals[1].pred, "student")
        self.assertEqual(rule.goals[1].args, ["y"])

    def test_initialize_bad_rule_1(self):
        with self.assertRaises(Exception):
            rule = Rule("")

    def test_initialize_bad_rule_2(self):
        with self.assertRaises(Exception):
            rule = Rule("advises(x,y):!-paper(v)")
